# Copyright (C) 2007 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Base application class for the InteProxy."""

from inteproxy.httpserver import ServerThread


class InteProxyApplication(object):

    """Base Application object

    An InteProxyApplication instance servers as the basis of the user
    interface of InteProxy.  This base class implements some basic
    functionality needed by all InteProxy application objects.  Derived
    classes provide a concrete user interface with e.g. a real GUI or a
    mock interface for tests.

    Each InteProxyApplication has a server object that implements the
    actual proxy.  It can be any server object compatible with
    inteproxy.httpserver.ServerThread, but it will usually be an
    inteproxy.proxycore.MasterWorkerServer instance.
    InteProxyApplication can can run this server in a separate thread.
    """

    def __init__(self, server):
        """Inititialize the InteProxyApplication with the given server object"""
        self.server = server
        self.thread = None

    def start_inte_proxy_thread(self, daemon):
        """Start a new thread running the proxy

        If the parameter daemon is true, the new thread will be daemon
        thread (see Python's threading module for details on deamon
        threads).
        """
        self.thread = ServerThread(self.server)
        self.thread.start(daemon=daemon)

    def stop_inte_proxy_thread(self):
        """Stops a server thread previously started with start_inte_proxy_thread
        """
        self.thread.stop()

    def run(self):
        """Runs the InteProxy application.

        The default implementation simply calls the server's
        serve_forever method.  Derived classes may do something more
        sophisticated such as running the server in a separate thread
        and a GUI in the main thread.
        """
        self.server.serve_forever()

    def get_password(self, title):
        """Callback to ask for a username and password.

        The parameter title should be a string with a title for the
        password dialog.  The title should contain all information the
        user may need to determine which username and password they have
        to enter.  This method returns the entered username/password
        pair as a tuple.  If the user cancels the interaction, the
        return value is a tuple of two None objects.

        This method should be called by request handlers of the server
        when they need to query the user for usernames and passwords.

        This method has to be implemented by derived classes.  The
        default implementation only raises NotImplementedError.
        """
        raise NotImplementedError

    def run_fees_dialog(self, title, fees, access_constraints):
        """Callback to run the fees dialog.

        The parameters are strings with the title of the dialog and the
        fees and access constraints information taken from a WMS
        server's GetCapabilities response.

        This method has to be implemented by derived classes.  The
        default implementation only raises NotImplementedError.
        """
        raise NotImplementedError
