# Copyright (C) 2008 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Abstractions for http request and response messages"""

from StringIO import StringIO

class HTTPMessage(object):

    """Base class for HTTP Messages in InteProxy.

    Parameters
      headers -- The headers of the message as a mimetools.Message
                 instance.

      infile -- File-object like object from which the body of the
                message can be read.

    The headers object can be accessed as the headers attribute.  The
    body should be accessed through the body property or the message's
    read method.  If the body is read via the body property, the body is
    read into memory from the infile once.  The body can also be set by
    assignment ot the body attribute or the set_body method.  The latter
    can also set the content type.

    If the body is read or set, the headers are adjusted a little.  The
    Content-length is set to the length of the body and any
    Transfer-encoding headers are removed.
    """

    def __init__(self, headers, infile):
        self.headers = headers
        self.infile = infile
        self._body = None
        self._body_stream = None

    def debug_log_message(self, log_function):
        for header, value in self.headers.items():
            log_function("header: %s:%r", header, value)
        if self.body_has_been_read():
            if self._body:
                log_function("body: %r", self._body)
            else:
                log_function("no body")
        else:
            log_function("body may be present but has not been read yet")

    def body_has_been_read(self):
        return self._body != None

    def set_body(self, body, content_type=None):
        self.headers["Content-length"] = str(len(body))
        del self.headers["Transfer-encoding"]
        if content_type is not None:
            self.headers["Content-type"] = content_type
        self._body = body

    def get_body(self):
        self.read_entire_message()
        return self._body

    body = property(get_body, set_body)

    def read_entire_message(self):
        raise NotImplementedError

    def read(self, amount):
        if self._body_stream is None and self.body_has_been_read():
            self._body_stream = StringIO(self.body)
        if self._body_stream is not None:
            return self._body_stream.read(amount)
        else:
            return self.infile.read(amount)


class HTTPRequestMessage(HTTPMessage):

    """Represents a HTTP Message for a request.

    Additional parameters:
       method -- The HTTP method of the request (usually 'GET' or 'POST'()

       uri -- The uri of the request

       version -- The http protocol version given in the request

    All of these parameters are available as attributes of the same
    name.
    """

    def __init__(self, method, uri, version, headers, infile):
        super(HTTPRequestMessage, self).__init__(headers, infile)
        self.method = method
        self.uri = uri
        self.version = version

    def debug_log_message(self, log_function):
        log_function("HTTPRequestMessage: %s %s %s",
                     self.method, self.uri, self.version)
        super(HTTPRequestMessage, self).debug_log_message(log_function)

    def read_entire_message(self):
        if self.body_has_been_read():
            return
        body = None
        length = int(self.headers.get("Content-Length", "0"))
        if length:
            # FIXME: check whether length bytes were read
            body = self.infile.read(length)

        if body is not None:
            self.set_body(body)


class HTTPResponseMessage(HTTPMessage):

    """Represents a HTTP Message for a response.

    Additional parameters:
       version -- The http protocol version given in the response

       status -- The HTTP status of the response

       reason -- The reason given in the response

    All of these parameters are available as attributes of the same
    name.
    """

    def __init__(self, version, status, reason, headers, infile):
        super(HTTPResponseMessage, self).__init__(headers, infile)
        self.version = version
        self.status = status
        self.reason = reason

    def debug_log_message(self, log_function):
        log_function("HTTPResponseMessage: %s %s %s",
                     self.version, self.status, self.reason)
        super(HTTPResponseMessage, self).debug_log_message(log_function)

    def read_entire_message(self):
        if self.body_has_been_read():
            return
        self.set_body(self.infile.read())
