#! /usr/bin/python
# Copyright (C) 2006, 2007, 2008 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""The HTTP proxy at the core of InteProxy"""

import sys
import httplib
import traceback
import time
import urlparse
import BaseHTTPServer
import socket

from inteproxy.threadpool import ThreadPool
from inteproxy.feesdialog import handle_fees_and_access_constraints
from inteproxy.httpserver import HTTPServer
from inteproxy.httpmessage import HTTPRequestMessage, HTTPResponseMessage
from inteproxy.proxyconnection import HTTPSProxyConnection


# same as the BaseHTTPRequestHandler method, but as a standalone function:
def log_date_time_string():
    """Return the current time formatted for logging."""
    now = time.time()
    year, month, day, hh, mm, ss, x, y, z = time.localtime(now)
    s = "%02d/%3s/%04d %02d:%02d:%02d" % (
        day, BaseHTTPServer.BaseHTTPRequestHandler.monthname[month],
        year, hh, mm, ss)
    return s

class InteProxyHTTPRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):

    # Whether debug output should be logged.  Supported values:
    #     0    no debug output
    #     1    normal debug level with abbreviated log of the data
    #          transferred
    #   >=2    Also logs all the data transferred by the proxy
    #
    debug_level = 0

    def do_GET(self):
        """Handle a GET request"""
        self.handle_proxy_request()

    def do_POST(self):
        """Handle a POST request"""
        self.handle_proxy_request()

    def handle_proxy_request(self):
        """Handle a request that needs to be 'proxied'.
        """
        self.log_debug("Got %s request for %s, version '%s'",
                    self.command, self.path, self.request_version)
        for header, value in self.headers.items():
            self.log_debug("header from client: %s:%r", header, value)

        #
        #  Read the client request
        #
        client_request = self.read_client_request()

        #
        # Determine the transcoder to use
        #
        transcoder = self.server.transcoder_map.get_transcoder(self.command,
                                                               self.path)

        #
        # convert request according to the transcoder rules
        #
        transcoder.convert_request(client_request)
        self.log_debug("modified client_request:")
        client_request.debug_log_message(self.log_debug)

        #
        # Create and send new request to the real remote host.
        #
        remote_url = transcoder.get_url()
        self.log_debug("Converted url: %r", remote_url)
        response = None
        try:
            response = self.open_http_connection(remote_url, client_request)
        except socket.error, exc:
            # some low lever error occurred (e.g. the connection to an
            # upstream proxy could not be established).  Send a Bad
            # Gateway error to the client
            self.send_error(502)

        if response is not None:
            response.debug_log_message(self.log_debug)
            # check for fees and access constraints and run a dialog
            handle_fees_and_access_constraints(remote_url, response)
            self.rewrite_urls(response)
            self.handle_response(response)

        self.log_debug("request finished")

    def read_client_request(self):
        """Read the client request including the body, if any.
        The request is returned as a TranscoderRequest instance."""
        return HTTPRequestMessage(self.command, self.path, self.request_version,
                                  self.headers, self.rfile)

    def open_http_connection(self, remote_url, client_request):
        """Open a HTTP connection to remote_url and send client_request

        This method handles both http and https URLs as well as proxies
        for both.  The return value is the httplib response object.
        """
        scheme, netloc, path, query, fragment = urlparse.urlsplit(remote_url)
        # the URI used in the request.  Usually, it's like remote_url
        # but with schem and netloc removed.
        request_uri = urlparse.urlunsplit(("", "", path, query, fragment))
        if scheme == "http":
            connection_class = httplib.HTTPConnection
            if self.server.http_proxy_url:
                netloc = urlparse.urlsplit(self.server.http_proxy_url)[1]
                request_uri = remote_url
        elif scheme == "https":
            connection_class = httplib.HTTPSConnection
            if self.server.https_proxy_url:
                netloc = urlparse.urlsplit(self.server.https_proxy_url)[1]
                request_uri = remote_url
                connection_class = HTTPSProxyConnection

        connection = connection_class(netloc)
        if self.request_version == "HTTP/1.0":
            connection._http_vsn = 10

        connection.putrequest(client_request.method, request_uri,
                              skip_accept_encoding=True)

        for header, value in client_request.headers.items():
            if header.lower() == "host":
                # the host header will be set by httplib and it should
                # not be set twice, so we omit it.  The value we
                # received from the client may have been wrong anyway as
                # in the old inteproxy scheme it refers to the host this
                # proxy is running on.
                continue
            else:
                connection.putheader(header, value)
        connection.endheaders()

        if client_request.body is not None:
            self.log_debug("Sending body to server: %r", client_request.body)
            connection.send(client_request.body)

        self.log_debug("request sent")

        response = connection.getresponse()
        for version_int, version in [(11, "HTTP/1.1"),
                                     (10, "HTTP/1.0"),
                                     ( 9, "HTTP/0.9")]:
            if response.version == version_int:
                break
        else:
            version = "UNKNOWN"
            self.log_debug("Unknown http version from server: %s",
                           response.version)
        if version != self.request_version:
            self.log_debug("Response version %r does not match"
                           " request version %r",
                           version, self.request_version)
        response_message = HTTPResponseMessage(version, response.status,
                                               response.reason,
                                               response.msg, response)

        return response_message


    def handle_response(self, response):
        # The HTTP version in the reply generated by send_response is
        # taken from self.protocol_version.  We simply set it to
        # whatever version was returned by the remote host.  That's what
        # the client would have received anyway, if it hadn't connected
        # via the InteProxy.
        self.protocol_version = response.version
        self.send_response(response.status, response.reason)

        for header, value in response.headers.items():
            self.log_debug("header to client: %s:%r", header, value)
            self.send_header(header, value)
        self.end_headers()

        transfer_encoding = response.headers.get("Transfer-encoding")
        self.transfer_data(response.read, self.wfile.write,
                           chunked = (transfer_encoding == "chunked"))

    def transfer_data(self, read, write, length=None, chunked=False):
        """Transfer data from one 'file' to another in chunks

        The files are given by their read and write methods so it
        doesn't have to be a file.  The read parameter must be callable
        with an integer argument indicating the maximum number of bytes
        to read and the write parameter must be callable with a string.
        If the parameter chunked is true, the method uses the 'chunked'
        transfer encoding when writing the data.
        """

        # wrap the read/write functions if debug logging is active so
        # that the data read from the server and written to the client
        # is logged.
        if self.debug_level > 0:
            orig_read = read
            orig_write = write

            def limit_length(data):
                """Returns a shortened version of the string data"""
                if self.debug_level < 2 and len(data) > 13:
                    data = "%s ... %x bytes ... %s" % (data[:5],
                                                       len(data) - 10,
                                                       data[-5:])
                return data

            def read(length):
                data = orig_read(length)
                self.log_debug("from server: %r", limit_length(data))
                return data
            def write(data):
                self.log_debug("to client: %r", limit_length(data))
                orig_write(data)

        # Now transfer the data in blocks of max_chunk_size
        max_chunk_size = 4096
        while 1:
            if length is not None:
                chunk_size = min(length, max_chunk_size)
            else:
                chunk_size = max_chunk_size
            chunk = read(chunk_size)
            if not chunk:
                break
            if length is not None:
                length -= len(chunk)
            if chunked:
                write("%x\r\n" % len(chunk))
            write(chunk)
            if chunked:
                write("\r\n")
        if chunked:
            write("0\r\n\r\n")

    def rewrite_urls(self, response):
        """Rewrites URLs in the response if enabled in the server

        This method rewrites URLs in the response if the request is a
        old-style InteProxy request -- one where InteProxy is not used
        as a normal HTTP proxy -- and rewriting has been activated in
        the server.  The actual rewriting is done by the server's
        transcoder_map.
        """
        if not urlparse.urlsplit(self.path)[0] and self.server.rewrite_urls:
            transcoder_map = self.server.transcoder_map
            prefix = self.server.get_inteproxy_url()
            response.body = transcoder_map.rewrite_urls(response.body, prefix,
                                                        self.log_debug)

    def log_exception(self, exc_info):
        """Log an exception

        This method produces a log message via self.log_message and then
        prints a the exception given by exc_info, which should be the
        return value of sys.exc_info, to stderr.  By default log_message
        also writes to stderr, so both the message and the traceback
        should go to the same place.
        """
        self.log_message("Exception for previous 500 code\n")
        traceback.print_exception(*(exc_info + (None, sys.stderr)))
        sys.stderr.flush()

    def log_debug(self, template, *args):
        if self.debug_level > 0:
            message = "DEBUG: " + template % args
            message = message.replace("%", "%%")
            self.log_message(message)


class MasterWorkerServer(HTTPServer):

    do_shutdown = False

    def __init__(self, server_address, RequestHandlerClass, num_workers,
                 http_proxy_url, https_proxy_url, transcoder_map,
                 rewrite_urls=False):
        HTTPServer.__init__(self, server_address, RequestHandlerClass)
        self.http_proxy_url = http_proxy_url
        self.https_proxy_url = https_proxy_url
        self.transcoder_map = transcoder_map
        self.rewrite_urls = rewrite_urls
        self.thread_pool = ThreadPool(num_workers, lambda f: f())
        sys.stderr.write("[%s] starting %d worker threads\n" \
                         % (log_date_time_string(), num_workers))
        self.thread_pool.start()

    def get_inteproxy_url(self):
        """Returns the URL of this InteProxy"""
        return "http://localhost:%d/" % self.getsockname()[1]

    def process_request(self, request, client_address):
        """Put the request into the queue to be handled by the worker thread
        """
        def process_in_worker():
            try:
                self.finish_request(request, client_address)
                self.close_request(request)
            except:
                self.handle_error(request, client_address)
                self.close_request(request)
        sys.stderr.write("[%s] queue contains %d items\n" %
                         (log_date_time_string(),
                          self.thread_pool.queue.qsize()))
        self.thread_pool.put(process_in_worker)

    def handle_error(self, request, client_address):
        """Override to integrate the error reporting better with the other logs
        """
        sys.stderr.write("[%s] Exception while handling request from %r\n"
                         % (log_date_time_string(), client_address))
        traceback.print_exception(*(sys.exc_info() + (None, sys.stderr)))
        sys.stderr.flush()

