# Copyright (C) 2007, 2008 by Intevation GmbH
# Authors:
# Bernhard Herzog <bh@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

"""Functional tests for InteProxy"""

import sys
import os
import httplib
import unittest

import serversupport

from inteproxy.proxycore import MasterWorkerServer, InteProxyHTTPRequestHandler
from inteproxy.transcoder import create_transcoder_map
import inteproxy.httpserver as httpserver



class ServerTest(unittest.TestCase):

    """Base class for tests that run the InteProxy in a thread"""


    # The content of the remote server in the format used by
    # serversupport.HTTPServer.  Override in derived classes.
    remote_contents = []

    # transcoders to use
    transcoder_definitions = []

    # whether the server should rewrite urls
    rewrite_urls = False

    def setUp(self):
        self.old_stderr = sys.stderr
        sys.stderr = open(os.path.devnull, "w")

        http_proxy = self.create_http_proxy()
        if http_proxy:
            self.http_proxy = httpserver.ServerThread(http_proxy)
            self.http_proxy.start(daemon=True)
            self.http_proxy_url = "http://localhost:%d/" \
                                  % (self.http_proxy.server_port,)
        else:
            self.http_proxy = self.http_proxy_url = None

        proxyserver = MasterWorkerServer(("127.0.0.1", 0),
                                         InteProxyHTTPRequestHandler, 5,
                                         self.http_proxy_url, None,
                                   transcoder_map=self.create_transcoder_map(),
                                         rewrite_urls=self.rewrite_urls)
        self.server = httpserver.ServerThread(proxyserver)
        self.server.start(daemon=True)
        remote_server = serversupport.HTTPServer(self.remote_contents)
        self.remote_server = httpserver.ServerThread(remote_server)
        self.remote_server.start(daemon=True)
        self.remote_server_base_url = "http://localhost:%d/" \
                                      % (self.remote_server.server_port,)

    def create_transcoder_map(self):
        transcoder_map = transcoder_map=create_transcoder_map()
        for host, path, cls in self.transcoder_definitions:
            transcoder_map.add_host(host, path, cls)
        return transcoder_map

    def create_http_proxy(self):
        """Override in derived classes to create a proxy upstream from inteproxy
        The method should return the proxy server object object.
        """
        return None

    def tearDown(self):
        try:
            for server in [self.remote_server, self.http_proxy,
                           self.server]:
                if server is not None:
                    server.stop()
        finally:
            sys.stderr = self.old_stderr



class TestInteProxy(ServerTest):

    remote_contents = [
        ("/wms", [("Content-Type", "text/plain")], "wms data"),
        ("/search?q=foo#section1", [("Content-Type", "text/plain")],
         "some text"),
        ]

    def test_httpproxy_passthrough(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET", self.remote_server_base_url + "wms")
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data, "wms data")

    def test_httpproxy_with_query_and_fragment(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     self.remote_server_base_url + "search?q=foo#section1")
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data, "some text")


class TestInteProxyURLRewriting(ServerTest):

    remote_contents = [
        ("/rewrite-01", [("Content-Type", "text/plain")],
         "An url that may be rewritten: http://example.com/a/wms"),
        ("/rewrite-02", [("Content-Type", "text/plain")],
         "An URL that may be rewritten: http://example.com/a/wms\n"
         "and one that may not: http://example.com/foo\n"
         "and one that may: http://example.com/another/wms\n"),
        ]


    transcoder_definitions = [
        ("example.com", "/a/wms", "owsproxy"),
        ("example.com", "/another/wms", "owsproxy"),
        ]

    rewrite_urls = True

    def test_httpproxy_url_rewriting_single_url(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-01"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An url that may be rewritten:"
                          " http://localhost:%d/example.com/a/wms"
                          % (self.server.server_port,))

    def test_httpproxy_url_rewriting_multiple_urls(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET",
                     "http://localhost:%d/localhost:%d/rewrite-02"
                     % (self.server.server_port,
                        self.remote_server.server_port))
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data,
                          "An URL that may be rewritten:"
                          " http://localhost:%d/example.com/a/wms\n"
                          "and one that may not: http://example.com/foo\n"
                          "and one that may:"
                          " http://localhost:%d/example.com/another/wms\n"
                          % (self.server.server_port,
                             self.server.server_port))



class TestInteProxyWithExtraProxy(ServerTest):

    remote_contents = [
        ("/wms", [("Content-Type", "text/plain")], "wms data"),
        ]

    def create_http_proxy(self):
        return MasterWorkerServer(("127.0.0.1", 0),
                                  InteProxyHTTPRequestHandler, 5,
                                  None, None,
                                  transcoder_map=create_transcoder_map())

    def test(self):
        http = httplib.HTTPConnection("localhost", self.server.server_port)
        http.request("GET", self.remote_server_base_url + "wms")
        response = http.getresponse()
        self.assertEquals(response.status, 200)
        data = response.read()
        self.assertEquals(data, "wms data")
